@@ -21,15 +21,27 @@
 """
 
 import struct
-from typing import Callable, Dict, List, Optional, Set, Tuple, cast
+import sys
+from typing import Any, Dict, List, Optional, Set, Tuple, Union
 
-from . import DNSMessage
-from .._dns import DNSAddress, DNSHinfo, DNSNsec, DNSPointer, DNSQuestion, DNSRecord, DNSService, DNSText
+from .._dns import (
+    DNSAddress,
+    DNSHinfo,
+    DNSNsec,
+    DNSPointer,
+    DNSQuestion,
+    DNSRecord,
+    DNSService,
+    DNSText,
+)
 from .._exceptions import IncomingDecodeError
-from .._logger import QuietLogger, log
+from .._logger import log
 from .._utils.time import current_time_millis
 from ..const import (
-    _TYPES,
+    _FLAGS_QR_MASK,
+    _FLAGS_QR_QUERY,
+    _FLAGS_QR_RESPONSE,
+    _FLAGS_TC,
     _TYPE_A,
     _TYPE_AAAA,
     _TYPE_CNAME,
@@ -38,6 +50,7 @@ from ..const import (
     _TYPE_PTR,
     _TYPE_SRV,
     _TYPE_TXT,
+    _TYPES,
 )
 
 DNS_COMPRESSION_HEADER_LEN = 1
@@ -47,20 +60,24 @@ MAX_NAME_LENGTH = 253
 
 DECODE_EXCEPTIONS = (IndexError, struct.error, IncomingDecodeError)
 
-UNPACK_3H = struct.Struct(b'!3H').unpack
-UNPACK_6H = struct.Struct(b'!6H').unpack
-UNPACK_HH = struct.Struct(b'!HH').unpack
-UNPACK_HHiH = struct.Struct(b'!HHiH').unpack
+UNPACK_3H = struct.Struct(b'!3H').unpack_from
+UNPACK_6H = struct.Struct(b'!6H').unpack_from
+UNPACK_HH = struct.Struct(b'!HH').unpack_from
+UNPACK_HHiH = struct.Struct(b'!HHiH').unpack_from
+
+_seen_logs: Dict[str, Union[int, tuple]] = {}
+_str = str
 
 
-class DNSIncoming(DNSMessage, QuietLogger):
-
+class DNSIncoming:
     """Object representation of an incoming DNS packet"""
 
     __slots__ = (
+        "_did_read_others",
+        'flags',
         'offset',
         'data',
-        'data_len',
+        '_data_len',
         'name_cache',
         'questions',
         '_answers',
@@ -83,10 +100,10 @@ class DNSIncoming(DNSMessage, QuietLogger):
         now: Optional[float] = None,
     ) -> None:
         """Constructor from string holding bytes of packet"""
-        super().__init__(0)
+        self.flags = 0
         self.offset = 0
         self.data = data
-        self.data_len = len(data)
+        self._data_len = len(data)
         self.name_cache: Dict[int, List[str]] = {}
         self.questions: List[DNSQuestion] = []
         self._answers: List[DNSRecord] = []
@@ -96,39 +113,72 @@ class DNSIncoming(DNSMessage, QuietLogger):
         self.num_authorities = 0
         self.num_additionals = 0
         self.valid = False
-        self._read_others = False
+        self._did_read_others = False
         self.now = now or current_time_millis()
         self.source = source
         self.scope_id = scope_id
-        self._parse_data(self._initial_parse)
-
-    def _initial_parse(self) -> None:
-        """Parse the data needed to initalize the packet object."""
-        self.read_header()
-        self.read_questions()
-        if not self.num_questions:
-            self.read_others()
-        self.valid = True
-
-    def _parse_data(self, parser_call: Callable) -> None:
-        """Parse part of the packet and catch exceptions."""
         try:
-            parser_call()
+            self._initial_parse()
         except DECODE_EXCEPTIONS:
-            self.log_exception_debug(
+            self._log_exception_debug(
                 'Received invalid packet from %s at offset %d while unpacking %r',
                 self.source,
                 self.offset,
                 self.data,
             )
 
+    def is_query(self) -> bool:
+        """Returns true if this is a query."""
+        return (self.flags & _FLAGS_QR_MASK) == _FLAGS_QR_QUERY
+
+    def is_response(self) -> bool:
+        """Returns true if this is a response."""
+        return (self.flags & _FLAGS_QR_MASK) == _FLAGS_QR_RESPONSE
+
+    @property
+    def truncated(self) -> bool:
+        """Returns true if this is a truncated."""
+        return (self.flags & _FLAGS_TC) == _FLAGS_TC
+
+    def _initial_parse(self) -> None:
+        """Parse the data needed to initalize the packet object."""
+        self._read_header()
+        self._read_questions()
+        if not self.num_questions:
+            self._read_others()
+        self.valid = True
+
+    @classmethod
+    def _log_exception_debug(cls, *logger_data: Any) -> None:
+        log_exc_info = False
+        exc_info = sys.exc_info()
+        exc_str = str(exc_info[1])
+        if exc_str not in _seen_logs:
+            # log the trace only on the first time
+            _seen_logs[exc_str] = exc_info
+            log_exc_info = True
+        log.debug(*(logger_data or ['Exception occurred']), exc_info=log_exc_info)
+
     @property
     def answers(self) -> List[DNSRecord]:
         """Answers in the packet."""
-        if not self._read_others:
-            self._parse_data(self.read_others)
+        if not self._did_read_others:
+            try:
+                self._read_others()
+            except DECODE_EXCEPTIONS:
+                self._log_exception_debug(
+                    'Received invalid packet from %s at offset %d while unpacking %r',
+                    self.source,
+                    self.offset,
+                    self.data,
+                )
         return self._answers
 
+    @property
+    def is_probe(self) -> bool:
+        """Returns true if this is a probe."""
+        return self.num_authorities > 0
+
     def __repr__(self) -> str:
         return '<DNSIncoming:{%s}>' % ', '.join(
             [
@@ -144,11 +194,7 @@ class DNSIncoming(DNSMessage, QuietLogger):
             ]
         )
 
-    def unpack(self, unpacker: Callable[[bytes], tuple], length: int) -> tuple:
-        self.offset += length
-        return unpacker(self.data[self.offset - length : self.offset])
-
-    def read_header(self) -> None:
+    def _read_header(self) -> None:
         """Reads header portion of packet"""
         (
             self.id,
@@ -157,38 +203,43 @@ class DNSIncoming(DNSMessage, QuietLogger):
             self.num_answers,
             self.num_authorities,
             self.num_additionals,
-        ) = self.unpack(UNPACK_6H, 12)
+        ) = UNPACK_6H(self.data)
+        self.offset += 12
 
-    def read_questions(self) -> None:
+    def _read_questions(self) -> None:
         """Reads questions section of packet"""
-        self.questions = [
-            DNSQuestion(self.read_name(), *self.unpack(UNPACK_HH, 4)) for _ in range(self.num_questions)
-        ]
+        for _ in range(self.num_questions):
+            name = self._read_name()
+            type_, class_ = UNPACK_HH(self.data, self.offset)
+            self.offset += 4
+            question = DNSQuestion(name, type_, class_)
+            self.questions.append(question)
 
-    def read_character_string(self) -> bytes:
+    def _read_character_string(self) -> bytes:
         """Reads a character string from the packet"""
         length = self.data[self.offset]
         self.offset += 1
-        return self.read_string(length)
+        return self._read_string(length)
 
-    def read_string(self, length: int) -> bytes:
+    def _read_string(self, length: int) -> bytes:
         """Reads a string of a given length from the packet"""
         info = self.data[self.offset : self.offset + length]
         self.offset += length
         return info
 
-    def read_others(self) -> None:
+    def _read_others(self) -> None:
         """Reads the answers, authorities and additionals section of the
         packet"""
-        self._read_others = True
+        self._did_read_others = True
         n = self.num_answers + self.num_authorities + self.num_additionals
         for _ in range(n):
-            domain = self.read_name()
-            type_, class_, ttl, length = self.unpack(UNPACK_HHiH, 10)
+            domain = self._read_name()
+            type_, class_, ttl, length = UNPACK_HHiH(self.data, self.offset)
+            self.offset += 10
             end = self.offset + length
             rec = None
             try:
-                rec = self.read_record(domain, type_, class_, ttl, length)
+                rec = self._read_record(domain, type_, class_, ttl, length)
             except DECODE_EXCEPTIONS:
                 # Skip records that fail to decode if we know the length
                 # If the packet is really corrupt read_name and the unpack
@@ -205,22 +256,30 @@ class DNSIncoming(DNSMessage, QuietLogger):
             if rec is not None:
                 self._answers.append(rec)
 
-    def read_record(self, domain: str, type_: int, class_: int, ttl: int, length: int) -> Optional[DNSRecord]:
+    def _read_record(
+        self, domain: _str, type_: int, class_: int, ttl: int, length: int
+    ) -> Optional[DNSRecord]:
         """Read known records types and skip unknown ones."""
         if type_ == _TYPE_A:
-            return DNSAddress(domain, type_, class_, ttl, self.read_string(4), created=self.now)
+            dns_address = DNSAddress(domain, type_, class_, ttl, self._read_string(4))
+            dns_address.created = self.now
+            return dns_address
         if type_ in (_TYPE_CNAME, _TYPE_PTR):
-            return DNSPointer(domain, type_, class_, ttl, self.read_name(), self.now)
+            return DNSPointer(domain, type_, class_, ttl, self._read_name(), self.now)
         if type_ == _TYPE_TXT:
-            return DNSText(domain, type_, class_, ttl, self.read_string(length), self.now)
+            return DNSText(domain, type_, class_, ttl, self._read_string(length), self.now)
         if type_ == _TYPE_SRV:
+            priority, weight, port = UNPACK_3H(self.data, self.offset)
+            self.offset += 6
             return DNSService(
                 domain,
                 type_,
                 class_,
                 ttl,
-                *cast(Tuple[int, int, int], self.unpack(UNPACK_3H, 6)),
-                self.read_name(),
+                priority,
+                weight,
+                port,
+                self._read_name(),
                 self.now,
             )
         if type_ == _TYPE_HINFO:
@@ -229,14 +288,15 @@ class DNSIncoming(DNSMessage, QuietLogger):
                 type_,
                 class_,
                 ttl,
-                self.read_character_string().decode('utf-8'),
-                self.read_character_string().decode('utf-8'),
+                self._read_character_string().decode('utf-8', 'replace'),
+                self._read_character_string().decode('utf-8', 'replace'),
                 self.now,
             )
         if type_ == _TYPE_AAAA:
-            return DNSAddress(
-                domain, type_, class_, ttl, self.read_string(16), created=self.now, scope_id=self.scope_id
-            )
+            dns_address = DNSAddress(domain, type_, class_, ttl, self._read_string(16))
+            dns_address.created = self.now
+            dns_address.scope_id = self.scope_id
+            return dns_address
         if type_ == _TYPE_NSEC:
             name_start = self.offset
             return DNSNsec(
@@ -244,8 +304,8 @@ class DNSIncoming(DNSMessage, QuietLogger):
                 type_,
                 class_,
                 ttl,
-                self.read_name(),
-                self.read_bitmap(name_start + length),
+                self._read_name(),
+                self._read_bitmap(name_start + length),
                 self.now,
             )
         # Try to ignore types we don't know about
@@ -254,20 +314,24 @@ class DNSIncoming(DNSMessage, QuietLogger):
         self.offset += length
         return None
 
-    def read_bitmap(self, end: int) -> List[int]:
+    def _read_bitmap(self, end: int) -> List[int]:
         """Reads an NSEC bitmap from the packet."""
         rdtypes = []
         while self.offset < end:
-            window = self.data[self.offset]
-            bitmap_length = self.data[self.offset + 1]
-            for i, byte in enumerate(self.data[self.offset + 2 : self.offset + 2 + bitmap_length]):
+            offset = self.offset
+            offset_plus_one = offset + 1
+            offset_plus_two = offset + 2
+            window = self.data[offset]
+            bitmap_length = self.data[offset_plus_one]
+            bitmap_end = offset_plus_two + bitmap_length
+            for i, byte in enumerate(self.data[offset_plus_two:bitmap_end]):
                 for bit in range(0, 8):
                     if byte & (0x80 >> bit):
                         rdtypes.append(bit + window * 256 + i * 8)
             self.offset += 2 + bitmap_length
         return rdtypes
 
-    def read_name(self) -> str:
+    def _read_name(self) -> str:
         """Reads a domain name from the packet."""
         labels: List[str] = []
         seen_pointers: Set[int] = set()
@@ -283,7 +347,7 @@ class DNSIncoming(DNSMessage, QuietLogger):
 
     def _decode_labels_at_offset(self, off: int, labels: List[str], seen_pointers: Set[int]) -> int:
         # This is a tight loop that is called frequently, small optimizations can make a difference.
-        while off < self.data_len:
+        while off < self._data_len:
             length = self.data[off]
             if length == 0:
                 return off + DNS_COMPRESSION_HEADER_LEN
@@ -300,8 +364,10 @@ class DNSIncoming(DNSMessage, QuietLogger):
                 )
 
             # We have a DNS compression pointer
-            link = (length & 0x3F) * 256 + self.data[off + 1]
-            if link > self.data_len:
+            link_data = self.data[off + 1]
+            link = (length & 0x3F) * 256 + link_data
+            lint_int = int(link)
+            if link > self._data_len:
                 raise IncomingDecodeError(
                     f"DNS compression pointer at {off} points to {link} beyond packet from {self.source}"
                 )
@@ -309,15 +375,16 @@ class DNSIncoming(DNSMessage, QuietLogger):
                 raise IncomingDecodeError(
                     f"DNS compression pointer at {off} points to itself from {self.source}"
                 )
-            if link in seen_pointers:
+            if lint_int in seen_pointers:
                 raise IncomingDecodeError(
                     f"DNS compression pointer at {off} was seen again from {self.source}"
                 )
-            linked_labels = self.name_cache.get(link, [])
+            linked_labels = self.name_cache.get(lint_int)
             if not linked_labels:
-                seen_pointers.add(link)
+                linked_labels = []
+                seen_pointers.add(lint_int)
                 self._decode_labels_at_offset(link, linked_labels, seen_pointers)
-                self.name_cache[link] = linked_labels
+                self.name_cache[lint_int] = linked_labels
             labels.extend(linked_labels)
             if len(labels) > MAX_DNS_LABELS:
                 raise IncomingDecodeError(
@@ -325,4 +392,4 @@ class DNSIncoming(DNSMessage, QuietLogger):
                 )
             return off + DNS_COMPRESSION_POINTER_LEN
 
-        raise IncomingDecodeError("Corrupt packet received while decoding name from {self.source}")
+        raise IncomingDecodeError(f"Corrupt packet received while decoding name from {self.source}")
